/*
 * This file has been donated to Jam.
 */

# include "jam.h"
# include "lists.h"
# include "parse.h"
# include "rules.h"
# include "regexp.h"
# include "headers.h"
# include "newstr.h"
# include "hash.h"
# include "hcache.h"

#ifdef OPT_HEADER_CACHE_EXT

/*
 * Craig W. McPheeters, Alias|Wavefront.
 *
 * hcache.c hcache.h - handle cacheing of #includes in source files
 *
 * Create a cache of files scanned for headers.  When starting jam,
 * look for the cache file and load it if present.  When finished the
 * binding phase, create a new header cache.  The cache contains
 * files, their timestamps and the header files found in their scan.
 * During the binding phase of jam, look in the header cache first for
 * the headers contained in a file.  If the cache is present and
 * valid, use its contents.  This results in dramatic speedups with
 * large projects (eg. 3min -> 1min startup for one project.)
 *
 * External routines:
 *    hcache_init() - read and parse the local .jamdeps file.
 *    hcache_done() - write a new .jamdeps file
 *    hcache() - return list of headers on target.  Use cache or do a scan.
 *    
 * The dependency file format is an ascii file with 1 line per target.
 * Each line has the following fields:
 * @boundname@ timestamp @file@ @file@ @file@ ... \n
 * */

struct hcachedata {
    char		*boundname;
    time_t		time;
    LIST		*includes;
    struct hcachedata	*next;
} ;

typedef struct hcachedata HCACHEDATA ;


static struct hash *hcachehash = 0;
static HCACHEDATA  *hcachelist = 0; 


#define CACHENAME	".jamhcache"


void
hcache_init()
{
    FILE	*f;
    HCACHEDATA  cachedata, *c;
    int		i, len;
    char	buf[20000];  /* Beware...fixed size buffer */
    char	*name;

    hcachehash = hashinit (sizeof (HCACHEDATA), "hcache");

    if (! (f = fopen (CACHENAME, "r" )))
	return;

    while (fgets (buf, sizeof (buf), f))
    {
	/*
	 * Ensure the entire line is there.  If jam is interrupted when 
	 * writing the cache file, it may have left a partial line.  Make
	 * sure the trailing \n is present.
	 */
	len = strlen (buf);
	if (buf[0] == '@' && buf[len-1] == '\n')
	{
	    i = 1;
	    while (buf[i] != '@') ++i;  /* go to end of name */
	    buf[i++] = 0;		/* terminate and skip */
	    ++i;			/* skip space */
	    c = &cachedata;
	    c->boundname = &(buf[1]);
	    if (hashenter (hcachehash, (HASHDATA **)&c))
	    {
		c->boundname = newstr (c->boundname);
		c->time = atoi (&(buf[i]));
		c->includes = 0;
		c->next = hcachelist;
		hcachelist = c;

		while (buf[i] && buf[i] != '@') ++i;	/* skip to first name */
		while (buf[i] == '@')
		{
		    ++i;
		    name = &buf[i];
		    while (buf[i] != '@') ++i;  /* go to end of name */
		    buf[i++] = 0;		/* terminate and skip */

		    c->includes = list_new (c->includes, newstr (name));

		    while (buf[i] && buf[i] != '@') ++i;   /* next name */
		}
	    }
	}
    }

    fclose (f);
}

void
hcache_done()
{
    FILE	*f;
    LIST	*l;
    HCACHEDATA  *c;

    if (!hcachehash)
	return;

    if (! (f = fopen (CACHENAME, "w" )))
	return;

    c = hcachelist;
    while (c)
    {
	fprintf (f, "@%s@ %d", c->boundname, c->time);
	for (l = c->includes; l; l = list_next(l))
	    fprintf (f, " @%s@", l->string);
	fprintf (f, "\n");
	c = c->next;
    }

    fclose (f);
}

LIST *
hcache (TARGET *t, int rec, regexp *re[])
{
    HCACHEDATA  cachedata, *c = &cachedata;
    LIST 	*l = 0;

    c->boundname = t->boundname;
    if (hashcheck (hcachehash, (HASHDATA **) &c))
    {
	if (c->time == t->time)
	{
	    if (DEBUG_HEADER)
		printf ("using header cache for %s\n", t->boundname);
	    l = list_copy (0, c->includes);
	    return l;
	}
	else {
	    if (DEBUG_HEADER)
	        printf ("header cache out of date for %s\n", t->boundname);
	    list_free (c->includes);
	    c->includes = 0;
	}
    }
    else {
	if (hashenter (hcachehash, (HASHDATA **)&c))
	{
	    c->boundname = newstr (c->boundname);
	    c->next = hcachelist;
	    hcachelist = c;
	}
    }

    /* 'c' points at the cache entry.  Its out of date. */

    l = headers1 (0, t->boundname, rec, re);

    c->time = t->time;
    c->includes = list_copy (0, l);

    return l;
}

#endif
